@@ -31,8 +31,8 @@
 
 #include "dfu.h"
 #include "usb_dfu.h"
-#include "sam7dfu.h"
-#include "dfu-version.h"
+#include "dfu_load.h"
+#include "quirks.h"
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
@@ -102,7 +102,7 @@ static int find_dfu_if(struct usb_device *dev, int (*handler)(struct dfu_if *, v
 	int rc;
 
 	memset(dfu_if, 0, sizeof(*dfu_if));
-	
+
 	for (cfg_idx = 0; cfg_idx < dev->descriptor.bNumConfigurations;
 	     cfg_idx++) {
 		cfg = &dev->config[cfg_idx];
@@ -128,16 +128,14 @@ static int find_dfu_if(struct usb_device *dev, int (*handler)(struct dfu_if *, v
 					dfu_if->product =
 						dev->descriptor.idProduct;
 					dfu_if->configuration = cfg_idx;
-					dfu_if->interface = 
+					dfu_if->interface =
 						intf->bInterfaceNumber;
-					dfu_if->altsetting = 
+					dfu_if->altsetting =
 						intf->bAlternateSetting;
 					if (intf->bInterfaceProtocol == 2)
-						dfu_if->flags |= 
-							DFU_IFF_DFU;
+						dfu_if->flags |= DFU_IFF_DFU;
 					else
-						dfu_if->flags &=
-							~DFU_IFF_DFU;
+						dfu_if->flags &= ~DFU_IFF_DFU;
 					if (!handler)
 						return 1;
 					rc = handler(dfu_if, v);
@@ -248,12 +246,12 @@ static int iterate_dfu_devices(struct dfu_if *dif,
 
 			if (dif && (dif->flags &
 			    (DFU_IFF_VENDOR|DFU_IFF_PRODUCT)) &&
-		    	    (dev->descriptor.idVendor != dif->vendor ||
-		     	    dev->descriptor.idProduct != dif->product))
+			    (dev->descriptor.idVendor != dif->vendor ||
+			    dev->descriptor.idProduct != dif->product))
 				continue;
 			if (dif && (dif->flags & DFU_IFF_DEVNUM) &&
-		    	    (atoi(usb_bus->dirname) != dif->bus ||
-		     	    dev->devnum != dif->devnum))
+			    (atoi(usb_bus->dirname) != dif->bus ||
+			    dev->devnum != dif->devnum))
 				continue;
 			if (!count_dfu_interfaces(dev))
 				continue;
@@ -368,6 +366,40 @@ static int resolve_device_path(struct dfu_if *dif)
 
 #endif /* !HAVE_USBPATH_H */
 
+/* Look for descriptor in the configuration descriptor output */
+static int usb_get_extra_descriptor(usb_dev_handle *udev, unsigned char type,
+			unsigned char index, void *resbuf, int size)
+{
+	char *cbuf;
+	int desclen, conflen, smallest;
+	int ret;
+	int p = 0;
+	int foundlen = 0;
+
+	conflen = (usb_device(udev))->config->wTotalLength;
+	cbuf = malloc(conflen);
+	ret = usb_get_descriptor(udev, USB_DT_CONFIG, index, cbuf, conflen);
+	if (ret < conflen) {
+		fprintf(stderr, "Warning: failed to retrieve complete"
+			"configuration descriptor\n");
+		conflen = ret;
+	}
+	while (p + 1 < conflen) {
+		desclen = (int) cbuf[p];
+		if (cbuf[p + 1] == type) {
+			smallest = desclen < size ? desclen : size;
+			memcpy(resbuf, &cbuf[p], smallest);
+			foundlen = smallest;
+			break;
+		}
+		p += desclen;
+	}
+	free(cbuf);
+	if (foundlen > 1)
+		return foundlen;
+	/* try to retrieve it through usb_get_descriptor directly */
+	return usb_get_descriptor(udev, type, index, resbuf, size);
+}
 
 static void help(void)
 {
@@ -390,7 +422,7 @@ static void help(void)
 
 static void print_version(void)
 {
-	printf("dfu-util version %s\n", VERSION "+svn" DFU_UTIL_VERSION);
+	printf("dfu-util version %s\n", VERSION);
 }
 
 static struct option opts[] = {
@@ -425,6 +457,8 @@ int main(int argc, char **argv)
 	int num_devs;
 	int num_ifs;
 	unsigned int transfer_size = 0;
+	unsigned int default_transfer_size = 1024;
+	unsigned int host_page_size;
 	enum mode mode = MODE_NONE;
 	struct dfu_status status;
 	struct usb_dfu_func_descriptor func_dfu;
@@ -432,14 +466,14 @@ int main(int argc, char **argv)
 	char *alt_name = NULL; /* query alt name if non-NULL */
 	char *end;
 	int final_reset = 0;
-	int page_size = getpagesize();
 	int ret;
-	
-	printf("dfu-util - (C) 2007-2008 by OpenMoko Inc.\n"
+
+	printf("dfu-util - (C) 2005-2008 by Weston Schmidt, Harald Welte and OpenMoko Inc.\n"
 	       "This program is Free Software and has ABSOLUTELY NO WARRANTY\n\n");
 
 	printf("dfu-util does currently only support DFU version 1.0\n\n");
 
+	host_page_size = getpagesize();
 	memset(dif, 0, sizeof(*dif));
 
 	usb_init();
@@ -569,7 +603,7 @@ int main(int argc, char **argv)
 
 	/* We have exactly one device. Its usb_device is now in dif->dev */
 
-	printf("Opening USB Device 0x%04x:0x%04x...\n", dif->vendor, dif->product);
+	printf("Opening DFU USB device... ");
 	dif->dev_handle = usb_open(dif->dev);
 	if (!dif->dev_handle) {
 		fprintf(stderr, "Cannot open device: %s\n", usb_strerror());
@@ -581,9 +615,14 @@ int main(int argc, char **argv)
 	if (!get_first_dfu_if(&_rt_dif))
 		exit(1);
 
+	printf("ID %04x:%04x\n", _rt_dif.vendor, _rt_dif.product);
+
+	/* find set of quirks for this device */
+	set_quirks(_rt_dif.vendor, _rt_dif.product);
+
 	if (!_rt_dif.flags & DFU_IFF_DFU) {
 		/* In the 'first round' during runtime mode, there can only be one
-	 	* DFU Interface descriptor according to the DFU Spec. */
+		* DFU Interface descriptor according to the DFU Spec. */
 
 		/* FIXME: check if the selected device really has only one */
 
@@ -609,6 +648,8 @@ int main(int argc, char **argv)
 		}
 		printf("state = %s, status = %d\n", 
 		       dfu_state_to_string(status.bState), status.bStatus);
+		if (!(quirks & QUIRK_POLLTIMEOUT))
+			usleep(status.bwPollTimeout * 1000);
 
 		switch (status.bState) {
 		case DFU_STATE_appIDLE:
@@ -757,6 +798,8 @@ status_again:
 	}
 	printf("state = %s, status = %d\n",
 	       dfu_state_to_string(status.bState), status.bStatus);
+	if (!(quirks & QUIRK_POLLTIMEOUT))
+		usleep(status.bwPollTimeout * 1000);
 
 	switch (status.bState) {
 	case DFU_STATE_appIDLE:
@@ -790,21 +833,33 @@ status_again:
 
 	if (!transfer_size) {
 		/* Obtain DFU functional descriptor */
-		ret = usb_get_descriptor(dif->dev_handle, 0x21, dif->interface,
-					 &func_dfu, sizeof(func_dfu));
+		ret = usb_get_extra_descriptor(dif->dev_handle, USB_DT_DFU,
+				dif->interface, &func_dfu, sizeof(func_dfu));
 		if (ret < 0) {
 			fprintf(stderr, "Error obtaining DFU functional "
 				"descriptor: %s\n", usb_strerror());
-			transfer_size = page_size;
 		} else {
 			transfer_size = get_int16_le(&func_dfu.wTransferSize);
+			printf("Device returned transfer size %i\n",
+				transfer_size);
 		}
 	}
-
-	if (transfer_size > page_size)
-		transfer_size = page_size;
-	
-	printf("Transfer Size = 0x%04x\n", transfer_size);
+	/* if returned zero or not detected (and not user specified) */
+	if (!transfer_size) {
+		transfer_size = default_transfer_size;
+		printf("Warning: Trying default transfer size %i\n",
+			transfer_size);
+	}
+	/* limitation of Linux usbdevio */
+	if (transfer_size > host_page_size) {
+		transfer_size = host_page_size;
+		printf("Limited transfer size to %i\n", transfer_size);
+	}
+	/* DFU specification */
+	if (transfer_size < dif->dev->descriptor.bMaxPacketSize0) {
+		transfer_size = dif->dev->descriptor.bMaxPacketSize0;
+		printf("Adjusted transfer size to %i\n", transfer_size);
+	}
 
 	if (DFU_STATUS_OK != status.bStatus ) {
 		printf("WARNING: DFU Status: '%s'\n",
@@ -817,16 +872,18 @@ status_again:
 			fprintf(stderr, "Error: %d\n", status.bStatus);
 			exit(1);
 		}
+		if (!(quirks & QUIRK_POLLTIMEOUT))
+			usleep(status.bwPollTimeout * 1000);
         }
 
 	switch (mode) {
 	case MODE_UPLOAD:
-		if (sam7dfu_do_upload(dif->dev_handle, dif->interface,
+		if (dfuload_do_upload(dif->dev_handle, dif->interface,
 				  transfer_size, filename) < 0)
 			exit(1);
 		break;
 	case MODE_DOWNLOAD:
-		if (sam7dfu_do_dnload(dif->dev_handle, dif->interface,
+		if (dfuload_do_dnload(dif->dev_handle, dif->interface,
 				  transfer_size, filename) < 0)
 			exit(1);
 		break;
